/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.3 Automatisches Lernen
 * Ganglinien
 * Copyright (C) 2011 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllernen.gllernen;

import java.util.HashMap;

import com.bitctrl.i18n.Messages;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.iav.gllib.gllib.dav.GanglinieMQ;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;

/**
 * Stellt für jedes Approximationsverfahren den Lern-Aspekt zur Verfügung.
 *
 * @author BitCtrl Systems GmbH, Thomas Thierfelder
 */
public class GanglinienAspekte {

	/**
	 * Der Standardaspekt für alle Approximationsverfahren
	 * <code>asp.agregation5Minuten</code>
	 **/
	private static final String ASP_AGREGATION5_MINUTEN = "asp.agregation5Minuten"; //$NON-NLS-1$

	/**
	 * Kommandozeilenargument für den Aspekt der B-Spline-Approximation.
	 */
	private static final String ARG_BSPLINE = "-bSpline=" //$NON-NLS-1$
			+ ASP_AGREGATION5_MINUTEN;

	/**
	 * Kommandozeilenargument für den Aspekt der Cubic-Spline-Approximation.
	 */
	private static final String ARG_CUBIC_SPLINE = "-cSpline=" //$NON-NLS-1$
			+ ASP_AGREGATION5_MINUTEN;

	/**
	 * Kommandozeilenargument für den Aspekt der Polyline-Approximation.
	 */
	private static final String ARG_POLYLINE = "-polyline=" //$NON-NLS-1$
			+ ASP_AGREGATION5_MINUTEN;

	/** Mappt das Approximationsverfahren auf die PID des Aspekts. */
	private static final HashMap<Integer, String> approximationsVerfahrenZuAnalyseDatenAspektPid = new HashMap<>();

	/** Mappt das Approximationsverfahren auf den Aspekt. */
	private static final HashMap<Integer, Aspect> approximationsVerfahrenZuAnalyseDatenAspekt = new HashMap<>();

	/**
	 * Initisiert dieses statische Objekt mit den Parametern aus der
	 * Kommandozeile.
	 *
	 * @param argumentList
	 *            die Datenverteiler-Argumentliste, wie sie in
	 *            {@link StandardApplication#parseArguments(ArgumentList)}
	 *            uebergeben wird.
	 */
	public static final void initialisiere(final ArgumentList argumentList) {
		final String aspBSpline = argumentList.fetchArgument(ARG_BSPLINE)
				.asString();
		final String aspCubicSpline = argumentList
				.fetchArgument(ARG_CUBIC_SPLINE).asString();
		final String aspPolyline = argumentList.fetchArgument(ARG_POLYLINE)
				.asString();

		approximationsVerfahrenZuAnalyseDatenAspektPid
				.put(GanglinieMQ.APPROX_BSPLINE, aspBSpline);
		approximationsVerfahrenZuAnalyseDatenAspektPid
				.put(GanglinieMQ.APPROX_CUBICSPLINE, aspCubicSpline);
		approximationsVerfahrenZuAnalyseDatenAspektPid
				.put(GanglinieMQ.APPROX_POLYLINE, aspPolyline);
		approximationsVerfahrenZuAnalyseDatenAspektPid
				.put(GanglinieMQ.APPROX_UNBESTIMMT, ASP_AGREGATION5_MINUTEN);
	}

	/**
	 * Validiert die ueber die Kommandozeile uebergebenen Aspekte.
	 *
	 * @param dav
	 *            Datenverteiler-Verbindung.
	 */
	public static final void validiere(final ClientDavInterface dav) {
		for (final Integer verfahren : approximationsVerfahrenZuAnalyseDatenAspektPid
				.keySet()) {
			final String aspPid = approximationsVerfahrenZuAnalyseDatenAspektPid
					.get(verfahren);
			final Aspect asp = dav.getDataModel().getAspect(aspPid);
			if (asp == null) {
				throw new IllegalArgumentException(Messages.get(
						GlLernenMsg.IllegalArgumentExceptionAspektUngueltig,
						aspPid));
			}
			approximationsVerfahrenZuAnalyseDatenAspekt.put(verfahren, asp);
		}
	}

	/**
	 * Erfragt zu einem Approximationsverfahren den Aspekt, unter dem Daten
	 * gelernt werden sollen.
	 *
	 * @param verfahren
	 *            das Approximationsverfahren.<br>
	 *            z.B.: {@link GanglinieMQ#APPROX_BSPLINE},
	 *            {@link GanglinieMQ#APPROX_CUBICSPLINE},
	 *            {@link GanglinieMQ#APPROX_POLYLINE},
	 *            {@link GanglinieMQ#APPROX_UNBESTIMMT}
	 * @return der Aspekt, unter dem Daten gelernt werden sollen.
	 */
	public static final Aspect getAspect(final Integer verfahren) {
		return approximationsVerfahrenZuAnalyseDatenAspekt.get(verfahren);
	}

}
