/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.3 Automatisches Lernen
 * Ganglinien
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllernen.gllernen;

import java.lang.Thread.UncaughtExceptionHandler;
import java.util.ArrayList;
import java.util.List;

import javax.persistence.EntityManagerFactory;

import com.bitctrl.util.jar.JarTools;
import com.bitctrl.util.logging.LoggerTools;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.iav.gllib.gllib.modell.GanglinienobjektFactory;
import de.bsvrz.iav.gllib.gllib.modell.db.GlSpeicher;
import de.bsvrz.iav.gllib.gllib.speicher.GlSpeicherServer;
import de.bsvrz.iav.gllib.gllib.speicher.GlSpeicherUtil;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.application.StandardApplicationRunner;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.SystemObjekt;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;

/**
 * Enthält die main-Methode der Applikation.
 *
 * @author BitCtrl Systems GmbH, Schumann
 */
public final class GlLernen
		implements StandardApplication, UncaughtExceptionHandler {

	/**
	 * Einsprungmethode der Applikation <i>Automatisches Lernen Ganglinien</i>.
	 *
	 * @param args
	 *            die Kommandozeilenargumente.
	 * @see #parseArguments(ArgumentList)
	 */
	public static void main(final String[] args) {
		StandardApplicationRunner.run(new GlLernen(), args);
	}

	/** Der Logger der Klasse. */
	private Debug log;

	/**
	 * Die PID des Messquerschnitts, für den als einziges gelernt wird. Wenn
	 * {@code null}, dann wird für alle gelernt.
	 */
	private String pidMQ;

	/** Ganglinien-Datenbank-IP. Standard: localhost. */
	private String dbIp;

	/** Ganglinien-Datenbank-Port. Standard: 1527. */
	private String dbPort;

	/** Ganglinien-Datenbank-Nutzer. Standard: derby. */
	private String dbUser;

	/** Ganglinien-Datenbank Passwort fuer Nutzer. Standard: derby. */
	private String dbPass;

	/** Ganglinien-Datenbank-Verzeichnis. Standard: gldb. */
	private String dbVerzeichnis;

	/** Die maximale Anzahl gleichzeitiger Threads beim Lernen. */
	private static int maxThreads;

	@SuppressWarnings("unused")
	private GlSpeicherServer glSpeicherServer;

	/**
	 * Gibt die maximale Anzahl gleichzeitiger Threads beim Lernen zurück.
	 *
	 * @return die maximale Threadanzahl.
	 */
	public static int getMaxThreads() {
		return maxThreads;
	}

	/**
	 * Allgemeine Initialisierung der Applikation.
	 */
	private GlLernen() {
		JarTools.printVersionInfo(GlLernen.class);
	}

	/**
	 * Initialisiert die datenverteilerrelevanten Eigenschaften der Anwendung
	 * und startet alle Submodule.
	 */
	@Override
	public void initialize(final ClientDavInterface connection)
			throws Exception {
		MessageSender.getInstance()
				.setApplicationLabel(getClass().getSimpleName() + "_"
						+ connection.getLocalApplicationObject().getId());

		final ObjektFactory factory;
		final List<MessQuerschnittAllgemein> mqListe;

		GanglinienAspekte.validiere(connection);

		factory = ObjektFactory.getInstanz();
		factory.setVerbindung(connection);
		factory.registerStandardFactories();
		factory.registerFactory(new GanglinienobjektFactory());

		final EntityManagerFactory emf = GlSpeicherUtil.getGlSpeicherEmf(dbIp,
				dbPort, dbUser, dbPass, dbVerzeichnis);
		Verwaltung.getInstanz();

		// Server für Anfragen nach historische Ganglinien starten
		GlSpeicher.init(emf);
		glSpeicherServer = new GlSpeicherServer(
				ObjektFactory.getInstanz().getVerbindung(), emf);

		mqListe = new ArrayList<>();
		if (pidMQ == null) {
			for (final SystemObjekt so : factory
					.bestimmeModellobjekte("typ.messQuerschnittAllgemein")) { //$NON-NLS-1$
				if (so instanceof MessQuerschnittAllgemein) {
					mqListe.add((MessQuerschnittAllgemein) so);
				}
			}
		} else {
			mqListe.add(
					(MessQuerschnittAllgemein) factory.getModellobjekt(pidMQ));
		}

		Verwaltung.getInstanz().setMessquerschnitte(mqListe);
		LogTools.log(log, GlLernenMsg.InfoGlLernenBereit);
	}

	/**
	 * Neben den Kommandozeilenargumenten für den Datenverteiler werden folgende
	 * unterstützt.
	 * <ul>
	 * <li>{@code -MQ=} hiermit kann zu Testzwecken das Lernen auf einen
	 * einzelnen Messquerschnitt begrenzt werden (optional).</li>
	 * <li>{@code -maxThreads=} hiermit die maximale Anzahl gleichzeitiger
	 * Lernthreads vorgegeben werden (optional).</li>
	 * </ul>
	 */
	@Override
	public void parseArguments(final ArgumentList argumentList)
			throws Exception {
		Thread.setDefaultUncaughtExceptionHandler(this);
		log = Debug.getLogger();

		if (argumentList.hasArgument("-MQ")) { //$NON-NLS-1$
			pidMQ = argumentList.fetchArgument("-MQ=").asNonEmptyString(); //$NON-NLS-1$
		} else {
			pidMQ = null;
		}

		if (argumentList.hasArgument("-maxThreads")) { //$NON-NLS-1$
			maxThreads = Integer.valueOf(argumentList
					.fetchArgument("-maxThreads=").asNonEmptyString()); //$NON-NLS-1$
		} else {
			maxThreads = -1;
		}

		dbIp = argumentList.fetchArgument("-dbIp=localhost").asString(); //$NON-NLS-1$
		dbPort = argumentList.fetchArgument("-dbPort=1527").asString(); //$NON-NLS-1$
		dbUser = argumentList.fetchArgument("-dbUser=derby").asString(); //$NON-NLS-1$
		dbPass = argumentList.fetchArgument("-dbPass=derby").asString(); //$NON-NLS-1$
		dbVerzeichnis = argumentList.fetchArgument("-dbVerzeichnis=gldb") //$NON-NLS-1$
				.asString();

		GanglinienAspekte.initialisiere(argumentList);
	}

	/**
	 * Gibt die Exception im Log und als Betriebsmeldung aus und beendet
	 * anschließend die Applikation.
	 */
	@Override
	public void uncaughtException(final Thread t, final Throwable e) {
		LogTools.log(log, GlLernenMsg.ErrorUnbekannterFehler, t,
				LoggerTools.getStackTrace(e));
		System.exit(-1);
	}

}
