/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.3 Automatisches Lernen
 * Ganglinien
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllernen.gllernen;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import com.bitctrl.Constants;
import com.bitctrl.util.Interval;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.iav.gllib.gllib.GanglinienOperationen.PatternMatchingErgebnis;
import de.bsvrz.iav.gllib.gllib.dav.GanglinieMQ;
import de.bsvrz.iav.gllib.gllib.dav.GanglinienMQOperationen;
import de.bsvrz.iav.gllib.gllib.modell.db.GlSpeicher;
import de.bsvrz.iav.gllib.gllib.modell.parameter.PdGanglinienModellAutomatischesLernenEreignis;
import de.bsvrz.sys.funclib.bitctrl.daf.BetriebsmeldungDaten;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.kalender.objekte.EreignisTyp;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Lernt für den Zeitraum eines Tages eine Ganglinie.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
public class Lernen {

	/** Der Logger der Klasse. */
	private final Debug log = Debug.getLogger();

	/** Die Ausgangsdaten des Lernens. */
	private final Ausgangsdaten ausgangsdaten;

	/** Der maximale Vergleichsabstand für das Verschmelzen von Ganglinien. */
	private final int maxVergleichsAbstand;

	/** Der Ereignistyp, für den eine Ganglinien gelernt wird. */
	private EreignisTyp ereignisTyp;

	/**
	 * Der Index der historische Ganglinie, die zum Lernen/Verschmelzen in Frage
	 * kommt. Der Wert -1 signalisiert, dass keine passende historische
	 * Ganglinie gefunden werden konnte.
	 *
	 * FIXME Sollte das nicht ein "int" sein, wenn im Fehlerfall -1 geliefert
	 * wird?
	 */
	private Integer historischeGanglinieIndex = -1;

	/**
	 * Der Abstand der gefundenen historischen Ganglinie zur Analyseganglinie.
	 */
	private int historischeGanglinieAbstand = Integer.MAX_VALUE;

	/** Die gelernte Ganglinie. */
	private GanglinieMQ ganglinie;

	/**
	 * Initialisierung den Lernprozess.
	 *
	 * @param mq
	 *            der Messquerschnitt für den die Ganglinien gelernt werden.
	 * @param lernzeitraum
	 *            der Zeitraum in dem gelernt werden soll.
	 * @param maxVergleichsAbstand
	 *            der maximale Vergleichsabstand für das Verschmelzen von
	 *            Ganglinien.
	 */
	public Lernen(final MessQuerschnittAllgemein mq,
			final Interval lernzeitraum, final int maxVergleichsAbstand) {
		this.maxVergleichsAbstand = maxVergleichsAbstand;
		ausgangsdaten = new Ausgangsdaten(mq, lernzeitraum);
	}

	/**
	 * Führt das Lernen aus.
	 */
	public void lerne() {
		LogTools.log(log,
				new BetriebsmeldungDaten(
						ausgangsdaten.getMq().getSystemObject()),
				GlLernenMsg.InfoLernenStart, ausgangsdaten.getMq(),
				ausgangsdaten.getLernzeitraum());

		if (!ausgangsdaten.abrufen()) {
			LogTools.log(log,
					new BetriebsmeldungDaten(
							ausgangsdaten.getMq().getSystemObject()),
					GlLernenMsg.InfoLernenAbgebrochen, ausgangsdaten.getMq());
			return;
		}

		// Ereignistyp zum Lernen bestimmen
		ereignisTyp = bestimmeLernEreignisTyp(false);
		if (ereignisTyp == null) {
			LogTools.log(log,
					new BetriebsmeldungDaten(
							ausgangsdaten.getMq().getSystemObject()),
					GlLernenMsg.InfoLernenAbgebrochen, ausgangsdaten.getMq());
			return;
		}

		// Nun kann gelernt werden
		boolean gelernt = false;
		if ((ausgangsdaten.getParamEreignisTypLernen().get(ereignisTyp)
				.getGanglinienTyp() == GanglinieMQ.TYP_ADDITIV)
				|| (ausgangsdaten.getParamEreignisTypLernen().get(ereignisTyp)
						.getGanglinienTyp() == GanglinieMQ.TYP_MULTIPLIKATIV)) {
			gelernt = lerneRelativeGanglinie();
		}
		if ((ausgangsdaten.getParamEreignisTypLernen().get(ereignisTyp)
				.getGanglinienTyp() == GanglinieMQ.TYP_ABSOLUT) || !gelernt) {
			if (ausgangsdaten.getParamEreignisTypLernen().get(ereignisTyp)
					.getGanglinienTyp() != GanglinieMQ.TYP_ABSOLUT) {
				ereignisTyp = bestimmeLernEreignisTyp(true);
				if (ereignisTyp == null) {
					LogTools.log(log,
							new BetriebsmeldungDaten(
									ausgangsdaten.getMq().getSystemObject()),
							GlLernenMsg.InfoLernenAbgebrochen,
							ausgangsdaten.getMq());
					return;
				}
			}
			lerneAbsoluteGanglinie();
		}

		sichereGanglinien();
	}

	/**
	 * Sucht den Ereignistyp mit der höchsten Priorität. Der ausgewählte
	 * Ereignistyp besitzt optional eine Ausschlussliste. Liegen Ereignistypen
	 * aus dieser Ausschlussliste im Lernzyklus an, wird der Ereignistyp mit der
	 * nächst niedrigeren Priorität ausgewählt.
	 *
	 * @param nurAbsolute
	 *            wenn {@code true}, dann werden nur Ereignistypen
	 *            berücksichtigt, zu denen eine absolute Ganglinie gelernt
	 *            werden soll. Wenn {@code false}, dann wird nicht zwischen
	 *            absolut und relativ unteschieden.
	 * @return der Ereignistyp oder {@code null}, wenn keiner gefunden werden
	 *         kann.
	 */
	private EreignisTyp bestimmeLernEreignisTyp(final boolean nurAbsolute) {
		EreignisTyp typ;
		LinkedList<EreignisTyp> liste;

		liste = new LinkedList<>(ausgangsdaten.getEreignisTypen());

		typ = null;
		while ((typ == null) && (liste.size() > 0)) {
			PdGanglinienModellAutomatischesLernenEreignis.Daten param;

			// Ereignistyp mit der höchsten Priorität bestimmen
			typ = liste.poll();

			param = ausgangsdaten.getParamEreignisTypLernen().get(typ);

			if (nurAbsolute
					&& (param.getGanglinienTyp() != GanglinieMQ.TYP_ABSOLUT)) {
				// Keine Typ für absolute Ganglinie, ergo weitersuchen
				typ = null;
				continue;
			}

			// Prüfen ob Ereignistypen anliegen, die in Ausschlussliste stehen
			for (final EreignisTyp et : param.getAusschlussliste()) {
				if (ausgangsdaten.getEreignisTypen().contains(et)) {
					// Typ kann nicht verwendet werden, ergo weitersuchen
					typ = null;
					break;
				}
			}
		}

		if (typ != null) {
			LogTools.log(log,
					new BetriebsmeldungDaten(
							ausgangsdaten.getMq().getSystemObject()),
					GlLernenMsg.FineLerneEreignisTyp, ausgangsdaten.getMq(),
					typ);
		} else {
			LogTools.log(log,
					new BetriebsmeldungDaten(
							ausgangsdaten.getMq().getSystemObject()),
					GlLernenMsg.WarningKeinEreignistyp, ausgangsdaten.getMq());
		}
		return typ;
	}

	/**
	 * Lernt eine absolute Ganglinie.
	 */
	private void lerneAbsoluteGanglinie() {
		PdGanglinienModellAutomatischesLernenEreignis.Daten param;

		LogTools.log(log,
				new BetriebsmeldungDaten(
						ausgangsdaten.getMq().getSystemObject()),
				GlLernenMsg.InfoLerneAbsoluteGanglinieStart,
				ausgangsdaten.getMq());

		param = ausgangsdaten.getParamEreignisTypLernen().get(ereignisTyp);

		bestimmeAbsoluteGanglinieMitGeringstemAbstand(ereignisTyp);

		if (param.getMaxAbstand() < maxVergleichsAbstand) {
			LogTools.log(log,
					new BetriebsmeldungDaten(
							ausgangsdaten.getMq().getSystemObject()),
					GlLernenMsg.WarningTriggerUnplausibel,
					ausgangsdaten.getMq(), ereignisTyp);
		}

		if (historischeGanglinieAbstand <= param.getMaxAbstand()) {
			// Es wurde eine passende Ganglinien zum Verschmelzen gefunden und
			// der Abstand ist nicht zu groß
			long gewicht;
			GanglinieMQ historischeGl;

			historischeGl = ausgangsdaten.getHistorischeGanglinien()
					.get(historischeGanglinieIndex);

			// Gewicht für das Verschmelzen bestimmen
			gewicht = historischeGl.getAnzahlVerschmelzungen();
			if (gewicht > param.getMaxWichtungsfaktor()) {
				gewicht = param.getMaxWichtungsfaktor();
			}
			if (gewicht <= 0) {
				gewicht = 1;
			}
			LogTools.log(log,
					new BetriebsmeldungDaten(
							ausgangsdaten.getMq().getSystemObject()),
					GlLernenMsg.FineVerschmelzeGanglinie, ausgangsdaten.getMq(),
					gewicht);

			// Ganglinie veschmelzen
			ganglinie = GanglinienMQOperationen.verschmelze(
					ausgangsdaten.getArchivGanglinie(
							param.getDarstellungsverfahren()),
					historischeGl, gewicht);
			ausgangsdaten.getHistorischeGanglinien()
					.set(historischeGanglinieIndex, ganglinie);
		} else {
			// Neue absolute Ganglinie anlegen
			if (historischeGanglinieIndex >= 0) {
				LogTools.log(log,
						new BetriebsmeldungDaten(
								ausgangsdaten.getMq().getSystemObject()),
						GlLernenMsg.FineAbstandUeberschritten,
						ausgangsdaten.getMq());
			} else {
				LogTools.log(log,
						new BetriebsmeldungDaten(
								ausgangsdaten.getMq().getSystemObject()),
						GlLernenMsg.FineNeueGanglinie, ausgangsdaten.getMq());
			}

			ganglinie = ausgangsdaten
					.getArchivGanglinie(param.getDarstellungsverfahren());
			ausgangsdaten.getHistorischeGanglinien().add(ganglinie);
			if ((historischeGanglinieAbstand != Integer.MAX_VALUE)
					&& (historischeGanglinieAbstand > maxVergleichsAbstand)) {
				LogTools.log(log,
						new BetriebsmeldungDaten(
								ausgangsdaten.getMq().getSystemObject()),
						GlLernenMsg.WarningAbstandUeberschritten,
						ausgangsdaten.getMq(), historischeGanglinieAbstand,
						maxVergleichsAbstand);
			}
		}
	}

	/**
	 * Sucht die absolute Ganglinie mit dem geringsten komplexen Abstand zur
	 * Analyseganglinie. Es wird der Index der Ganglinie und deren Abstand zur
	 * Analyseganglinie bestimmt.
	 *
	 * @param typ
	 *            der Ereignistyp, zu dem eine absolute Ganglinie gesucht wird.
	 * @see #historischeGanglinieIndex
	 * @see #historischeGanglinieAbstand
	 */
	private void bestimmeAbsoluteGanglinieMitGeringstemAbstand(
			final EreignisTyp typ) {
		PdGanglinienModellAutomatischesLernenEreignis.Daten paramTypLernen;
		int index;
		int abstand;

		LogTools.log(log,
				new BetriebsmeldungDaten(
						ausgangsdaten.getMq().getSystemObject()),
				GlLernenMsg.FinerSucheHistorischeGanglinieMitGeringstenAbstand,
				ausgangsdaten.getMq(), typ);

		paramTypLernen = ausgangsdaten.getParamEreignisTypLernen().get(typ);
		assert paramTypLernen
				.getGanglinienTyp() == GanglinieMQ.TYP_ABSOLUT : "Für ein Tagesereignis darf nur eine absolute Ganglinie gelernt werden."; //$NON-NLS-1$

		// Index der historischen Ganglinie mit dem geringsten Abstand bestimmen
		abstand = Integer.MAX_VALUE;
		index = -1;
		for (int i = 0; i < ausgangsdaten.getHistorischeGanglinien()
				.size(); ++i) {
			int d;
			GanglinieMQ g;

			g = ausgangsdaten.getHistorischeGanglinien().get(i);
			if (g.getEreignisTyp().equals(typ)) {
				d = GanglinienMQOperationen.komplexerAbstand(
						ausgangsdaten.getArchivGanglinie(
								paramTypLernen.getDarstellungsverfahren()),
						g, paramTypLernen.getVergleichsSchrittweite());
				LogTools.log(log,
						new BetriebsmeldungDaten(
								ausgangsdaten.getMq().getSystemObject()),
						GlLernenMsg.FinerSucheHistorischeGanglinieMitGeringstenAbstandUntersuche,
						ausgangsdaten.getMq(), i, d);
				if (d < abstand) {
					abstand = d;
					index = i;
				}
			}
		}

		if (index >= 0) {
			LogTools.log(log,
					new BetriebsmeldungDaten(
							ausgangsdaten.getMq().getSystemObject()),
					GlLernenMsg.FineSucheHistorischeGanglinieMitGeringstenAbstandGefunden,
					ausgangsdaten.getMq(), index, abstand,
					GanglinienMQOperationen.formatierterText(ausgangsdaten
							.getHistorischeGanglinien().get(index)));
		} else {
			LogTools.log(log,
					new BetriebsmeldungDaten(
							ausgangsdaten.getMq().getSystemObject()),
					GlLernenMsg.FineSucheHistorischeGanglinieMitGeringstenAbstandNichtGefunden,
					ausgangsdaten.getMq());
		}

		historischeGanglinieIndex = index;
		historischeGanglinieAbstand = abstand;
	}

	/**
	 * Sichert die Ganglinienliste im Datenverteiler.
	 */
	private void sichereGanglinien() {
		final PdGanglinienModellAutomatischesLernenEreignis.Daten param;
		final long zeitstempel;
		final ClientDavInterface dav;

		dav = ObjektFactory.getInstanz().getVerbindung();

		zeitstempel = dav.getTime();
		param = ausgangsdaten.getParamEreignisTypLernen().get(ereignisTyp);

		// Eigenschaften ergänzen bzw. aktualisieren
		assert ganglinie != null : "Irgendeine Ganglinie muss erzeugt oder verschmolzen werden."; //$NON-NLS-1$
		assert ereignisTyp != null : "Der Ereignistyp muss ausgewählt sein."; //$NON-NLS-1$
		ganglinie.setEreignisTyp(ereignisTyp);
		ganglinie.setApproximationDaK(param.getDarstellungsverfahren());
		LogTools.log(log,
				new BetriebsmeldungDaten(
						ausgangsdaten.getMq().getSystemObject()),
				GlLernenMsg.FinerGelernteGanglinie, ausgangsdaten.getMq(),
				GanglinienMQOperationen.formatierterText(ganglinie));
		LogTools.log(log,
				new BetriebsmeldungDaten(
						ausgangsdaten.getMq().getSystemObject()),
				GlLernenMsg.FinestListeHistorischeGanglinien,
				ausgangsdaten.getMq(),
				ausgangsdaten.getHistorischeGanglinien());

		while (zaehleGanglinien(ausgangsdaten.getHistorischeGanglinien(),
				ereignisTyp) > param.getMaxGanglinien()) {
			// Die älteste Ganglinie für den Ereignistyp löschen
			long t = Long.MAX_VALUE;
			int index = -1;

			for (int i = 0; i < ausgangsdaten.getHistorischeGanglinien()
					.size(); i++) {
				final GanglinieMQ ganglinieMQ = ausgangsdaten
						.getHistorischeGanglinien().get(i);
				if (ereignisTyp.equals(ganglinieMQ.getEreignisTyp())) {
					if (t > ganglinieMQ.getLetzteVerschmelzung()) {
						t = ganglinieMQ.getLetzteVerschmelzung();
						index = i;
					}
				}
			}
			if (index >= 0) {
				LogTools.log(log,
						new BetriebsmeldungDaten(
								ausgangsdaten.getMq().getSystemObject()),
						GlLernenMsg.FineMaxAnzahlHistorischerGanglinienUeberschritten,
						ausgangsdaten.getMq(), ereignisTyp,
						GanglinienMQOperationen.formatierterText(ausgangsdaten
								.getHistorischeGanglinien().get(index)));
				ausgangsdaten.getHistorischeGanglinien().remove(index);
			}
		}

		// Ganglinienliste des MQ aktualiseren

		ausgangsdaten.getHistorischeGanglinien().setZeitstempel(zeitstempel);
		GlSpeicher.getInstanz().write(ausgangsdaten.getHistorischeGanglinien());

		/**
		 * Vorher:
		 */
		// final PdGanglinie pd = ausgangsdaten.getMq().getParameterDatensatz(
		// PdGanglinie.class);
		//
		// ausgangsdaten.getHistorischeGanglinien().setZeitstempel(zeitstempel);
		// try {
		// pd.anmeldenSender();
		// pd.sendeDaten(ausgangsdaten.getHistorischeGanglinien());
		// } catch (final DatensendeException ex) {
		// LogTools.log(log,
		// GlLernenMsg.ErrorHistorischeGanglinienNichtGesichert,
		// ausgangsdaten.getMq(), ex.getLocalizedMessage());
		// } catch (final AnmeldeException ex) {
		// LogTools.log(log,
		// GlLernenMsg.ErrorHistorischeGanglinienNichtGesichert,
		// ausgangsdaten.getMq(), ex.getLocalizedMessage());
		// }
		// dav.sleep(Constants.MILLIS_PER_SECOND);
		// pd.abmeldenSender();
	}

	/**
	 * Zählt die Ganglinien eines Ereignistyps in einer Ganglinienliste.
	 *
	 * @param ganglinien
	 *            eine Ganglinienliste.
	 * @param ereignisTyp
	 *            ein Ereignistyp.
	 * @return die Anzahl der Ganglinien zum Ereignistyp.
	 */
	private static int zaehleGanglinien(final List<GanglinieMQ> ganglinien,
			final EreignisTyp ereignisTyp) {
		int i;

		i = 0;
		for (final GanglinieMQ g : ganglinien) {
			if (ereignisTyp.equals(g.getEreignisTyp())) {
				++i;
			}
		}
		return i;
	}

	/**
	 * Lernt eine relative Ganglinie (additiv oder multiplikativ).
	 *
	 * @return {@code true}, wenn eine Ganglinie gelernt wurde.
	 */
	private boolean lerneRelativeGanglinie() {
		final PdGanglinienModellAutomatischesLernenEreignis.Daten param;
		final PdGanglinienModellAutomatischesLernenEreignis.Daten paramAbsolut;
		final EreignisTyp typAbsolut;
		final List<GanglinieMQ> listeAbsoluteGl, listeRelativeGl;
		final Map<Integer, Integer> matchingHilfe;
		final Map<Integer, Integer> historischeGlHilfe;
		final List<GanglinieMQ> matching;
		final PatternMatchingErgebnis<GanglinieMQ> matchingErg;

		LogTools.log(log,
				new BetriebsmeldungDaten(
						ausgangsdaten.getMq().getSystemObject()),
				GlLernenMsg.InfoLerneRelativeGanglinieStart,
				ausgangsdaten.getMq());

		typAbsolut = bestimmeBezugsereignis(ereignisTyp);
		param = ausgangsdaten.getParamEreignisTypLernen().get(ereignisTyp);
		paramAbsolut = ausgangsdaten.getParamEreignisTypLernen()
				.get(typAbsolut);

		if (typAbsolut != null) {
			LogTools.log(log,
					new BetriebsmeldungDaten(
							ausgangsdaten.getMq().getSystemObject()),
					GlLernenMsg.FinerLerneRelativeGanglinieBezugsereignis,
					ausgangsdaten.getMq(), typAbsolut);
		} else {
			LogTools.log(log,
					new BetriebsmeldungDaten(
							ausgangsdaten.getMq().getSystemObject()),
					GlLernenMsg.FinerLerneRelativeGanglinieKeinBezugsereignis,
					ausgangsdaten.getMq());
			return false;
		}

		// Bestimme alle historischen Ganglinien zum Bezugsereignis
		listeAbsoluteGl = new ArrayList<>();
		for (final GanglinieMQ g : ausgangsdaten.getHistorischeGanglinien()) {
			if (g.getEreignisTyp().equals(typAbsolut)) {
				listeAbsoluteGl.add(g);
			}
		}
		assert listeAbsoluteGl
				.size() > 0 : "Es müssen absolute Ganglinien existieren, da ein Bezugstyp existiert."; //$NON-NLS-1$

		// Ganglinien-Index -> Index in historischen Ganglinien
		historischeGlHilfe = new HashMap<>();

		// Bestimme alle historischen Ganglinien zum zu lernenden Ereignis
		listeRelativeGl = new ArrayList<>();
		for (int i = 0; i < ausgangsdaten.getHistorischeGanglinien()
				.size(); ++i) {
			final GanglinieMQ g;

			g = ausgangsdaten.getHistorischeGanglinien().get(i);
			if (g.getEreignisTyp().equals(ereignisTyp)) {
				listeRelativeGl.add(g);
				historischeGlHilfe.put(listeRelativeGl.size() - 1, i);
			}
		}

		// Matching-Index -> Ganglinien-Index
		matchingHilfe = new HashMap<>();

		// Pattern-Matching ausführen
		matching = new ArrayList<>();
		for (int i = 0; i < listeAbsoluteGl.size(); ++i) {
			GanglinieMQ absoluteGl;

			absoluteGl = listeAbsoluteGl.get(i);
			if (listeRelativeGl.isEmpty()) {
				matching.add(absoluteGl.clone());
				matchingHilfe.put(matching.size() - 1, -1);
			} else {
				for (int j = 0; j < listeRelativeGl.size(); ++j) {
					final GanglinieMQ relativeGl;
					GanglinieMQ g;

					relativeGl = listeRelativeGl.get(j);
					if (relativeGl.getTyp() == GanglinieMQ.TYP_ADDITIV) {
						g = GanglinienMQOperationen.addiere(absoluteGl,
								relativeGl);
					} else if (relativeGl
							.getTyp() == GanglinieMQ.TYP_MULTIPLIKATIV) {
						g = GanglinienMQOperationen.multipliziere(absoluteGl,
								relativeGl);
					} else {
						assert false : "Die Ganglinie zu " + ereignisTyp //$NON-NLS-1$
								+ " muss relativ sein, war aber " //$NON-NLS-1$
								+ relativeGl.getTyp();
						continue;
					}
					matching.add(g);
					matchingHilfe.put(matching.size() - 1, j);
				}
			}
		}
		matchingErg = GanglinienMQOperationen.patternMatching(
				ausgangsdaten
						.getArchivGanglinie(param.getDarstellungsverfahren()),
				matching,
				paramAbsolut.getMatchingIntervallVor()
						* Constants.MILLIS_PER_SECOND,
				paramAbsolut.getMatchingIntervallNach()
						* Constants.MILLIS_PER_SECOND,
				paramAbsolut.getMatchingSchrittweite()
						* Constants.MILLIS_PER_SECOND);
		LogTools.log(log,
				new BetriebsmeldungDaten(
						ausgangsdaten.getMq().getSystemObject()),
				GlLernenMsg.FineErgebnisPatternMatching, ausgangsdaten.getMq(),
				matchingErg);
		historischeGanglinieIndex = historischeGlHilfe
				.get(matchingHilfe.get(matchingErg.getIndex()));
		LogTools.log(log,
				new BetriebsmeldungDaten(
						ausgangsdaten.getMq().getSystemObject()),
				GlLernenMsg.FineAktualisiereGanglinie, ausgangsdaten.getMq(),
				historischeGanglinieIndex,
				ausgangsdaten.getHistorischeGanglinien()
						.get(historischeGanglinieIndex));

		// Die gelernte Ganglinie bestimmen
		ausgangsdaten.getArchivGanglinie(param.getDarstellungsverfahren())
				.setApproximationDaK(ausgangsdaten.getHistorischeGanglinien()
						.get(historischeGanglinieIndex).getApproximationDaK());
		switch (ausgangsdaten.getHistorischeGanglinien()
				.get(historischeGanglinieIndex).getTyp()) {
		case GanglinieMQ.TYP_ADDITIV:
			ganglinie = GanglinienMQOperationen.subtrahiere(
					ausgangsdaten.getArchivGanglinie(
							param.getDarstellungsverfahren()),
					matchingErg.getGanglinie());
			break;
		case GanglinieMQ.TYP_MULTIPLIKATIV:
			ganglinie = GanglinienMQOperationen.dividiere(
					ausgangsdaten.getArchivGanglinie(
							param.getDarstellungsverfahren()),
					matchingErg.getGanglinie());
			break;
		default:
			assert false : "Der Typ der zu lernenden Ganglinie muss relativ sein."; //$NON-NLS-1$
		}

		LogTools.log(log,
				new BetriebsmeldungDaten(
						ausgangsdaten.getMq().getSystemObject()),
				GlLernenMsg.FineLerneGanglinie, ausgangsdaten.getMq(),
				ganglinie);

		// Neue Ganglinie oder mit vorhandener verschmelzen?
		if (matchingErg.getAbstand() < param.getMaxMatchingFehler()) {
			// Es wurde eine passende Ganglinien zum Verschmelzen gefunden und
			// der Abstand ist nicht zu groß
			int gewicht;
			GanglinieMQ historischeGl;

			historischeGl = ausgangsdaten.getHistorischeGanglinien()
					.get(historischeGanglinieIndex);

			// Gewicht für das Verschmelzen bestimmen
			gewicht = (int) historischeGl.getAnzahlVerschmelzungen();
			if (gewicht > param.getMaxWichtungsfaktor()) {
				gewicht = param.getMaxWichtungsfaktor();
			}
			if (gewicht == 0) {
				gewicht = 1;
			}
			LogTools.log(log,
					new BetriebsmeldungDaten(
							ausgangsdaten.getMq().getSystemObject()),
					GlLernenMsg.FineVerschmelzeGanglinie, ausgangsdaten.getMq(),
					gewicht);

			// Ganglinie veschmelzen
			ganglinie = GanglinienMQOperationen.verschmelze(ganglinie,
					historischeGl, gewicht);
			ausgangsdaten.getHistorischeGanglinien()
					.set(historischeGanglinieIndex, ganglinie);
		} else {
			// Neue relative Ganglinie anlegen
			if (historischeGanglinieIndex != null) {
				LogTools.log(log,
						new BetriebsmeldungDaten(
								ausgangsdaten.getMq().getSystemObject()),
						GlLernenMsg.FineAbstandUeberschritten,
						ausgangsdaten.getMq());
			} else {
				LogTools.log(log, GlLernenMsg.FineNeueGanglinie,
						ausgangsdaten.getMq());
			}

			ausgangsdaten.getHistorischeGanglinien().add(ganglinie);
			if ((historischeGanglinieAbstand != Integer.MAX_VALUE)
					&& (historischeGanglinieAbstand > maxVergleichsAbstand)) {
				LogTools.log(log,
						new BetriebsmeldungDaten(
								ausgangsdaten.getMq().getSystemObject()),
						GlLernenMsg.WarningAbstandUeberschritten,
						ausgangsdaten.getMq(), historischeGanglinieAbstand,
						maxVergleichsAbstand);
			}
		}

		return true;
	}

	/**
	 * Bestimmt den Ereignistyp mit einer absoluten Ganglinie im Lernzyklus mit
	 * der höchsten Priorität auf den sich ein Ereignis einer relativen
	 * Ganglinie beziehen kann. Der ausgewählte Ereignistyp besitzt optional
	 * eine Ausschlussliste. Liegen Ereignistypen aus dieser Ausschlussliste im
	 * Lernzyklus an, wird der Ereignistyp mit der nächst niedrigeren Priorität
	 * ausgewält.
	 *
	 * @param relTyp
	 *            der relative Ereignistyp zu dem ein passender absoluter
	 *            Ereignistyp gesucht ist.
	 * @return der gefundene Ereignistyp oder {@code null}, wenn keiner bestimmt
	 *         werden konnte.
	 */
	private EreignisTyp bestimmeBezugsereignis(final EreignisTyp relTyp) {
		PdGanglinienModellAutomatischesLernenEreignis.Daten relTypParam;
		EreignisTyp typ;
		LinkedList<EreignisTyp> liste;

		liste = new LinkedList<>(ausgangsdaten.getEreignisTypen());

		relTypParam = ausgangsdaten.getParamEreignisTypLernen().get(relTyp);

		typ = null;
		while (liste.size() > 0) {
			PdGanglinienModellAutomatischesLernenEreignis.Daten param;

			// Ereignistyp mit der höchsten Priorität bestimmen
			typ = liste.poll();
			param = ausgangsdaten.getParamEreignisTypLernen().get(typ);

			// Tagesereignis gefunden?
			if (param.getGanglinienTyp() != GanglinieMQ.TYP_ABSOLUT) {
				typ = null;
				continue;
			}

			// Prüfen ob Ereignistypen anliegen, die in Ausschlussliste stehen.
			for (final EreignisTyp et : param.getAusschlussliste()) {
				if (ausgangsdaten.getEreignisTypen().contains(et)) {
					typ = null;
					continue;
				}
			}

			// Ist das Tagesereignis mit dem relativen Ereignis kombinierbar?
			final List<EreignisTyp> bezugsereignistypen = relTypParam
					.getBezugsereignistypen();
			if (!((bezugsereignistypen.isEmpty())
					|| (bezugsereignistypen.contains(typ)))) {
				typ = null;
				continue;
			}

			// Gibt es eine Ganglinie für das Tagesereignis?
			boolean gefunden = false;
			for (final GanglinieMQ g : ausgangsdaten
					.getHistorischeGanglinien()) {
				if (g.getEreignisTyp().equals(typ)) {
					gefunden = true;
					break;
				}
			}
			if (gefunden) {
				break;
			}
			typ = null;
		}

		return typ;
	}

}
