/*
 * Segment 5 Intelligente Analyseverfahren, SWE 5.3 Automatisches Lernen
 * Ganglinien
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.iav.gllernen.gllernen;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.junit.AfterClass;
import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Test;

import com.bitctrl.Constants;
import com.bitctrl.resource.Configuration;
import com.bitctrl.util.CronPattern;

import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.iav.gllib.gllib.dav.GanglinieMQ;
import de.bsvrz.iav.gllib.gllib.junit.ArchivdatenFactory;
import de.bsvrz.iav.gllib.gllib.junit.EreignisFactory;
import de.bsvrz.iav.gllib.gllib.junit.GanglinienFactory;
import de.bsvrz.iav.gllib.gllib.modell.GanglinienobjektFactory;
import de.bsvrz.iav.gllib.gllib.modell.db.GlSpeicher;
import de.bsvrz.iav.gllib.gllib.modell.parameter.PdGanglinienModellAutomatischesLernen;
import de.bsvrz.sys.funclib.bitctrl.modell.ObjektFactory;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;

/**
 * Führt den Prüffall 8 <em>Automatisches Lernen</em> aus. Es wird das Lernen
 * von absoluten und relativen Ganglinien getestet.
 * <p>
 * Es wird ein beliebiger Messquerschnitt für den Test benötigt. Das Archiv muss
 * laufen und darf für den Testmessquerschnitt keine Kurzzeitdaten beinhalten.
 * Der Ereigniskalender muss laufen.
 * <p>
 * <strong>Hinweis:</strong> Die SWE 5.3 <em>Automatisches Lernen</em> muss
 * laufen und bis heute vor 10 Tagen gelernt haben. Dazu muss vor dem Start der
 * SWE 5.3 der globale Lernparameter
 * {@code atg.ganglinienModellAutomatischesLernen} wie folgt eingestellt werden:
 * AlgAktualisierungsintervall: "* * * * *", AlgDatenMindestalter: 10&nbsp;Tage,
 * AlgMaxVergleichsAbstand: 100&nbsp;%.
 *
 * @author BitCtrl Systems GmbH, Falko Schumann
 */
@SuppressWarnings("nls")
public final class Prueffall8 {

	/** Der Logger der Klassse. */
	private static Logger log = Logger.getLogger(Prueffall8.class.getName());

	/** Die Factory für die Kalenderdaten des Tests. */
	private static EreignisFactory ereignisFactory;

	/** Die Factory für die Ganglinien des Tests. */
	private static GanglinienFactory ganglinienFactory;

	/** Der Messquerschnitt an dem der Test stattfindet. */
	private static MessQuerschnittAllgemein mq;

	/** Der Parameter, der das Lernen steuert. */
	private static PdGanglinienModellAutomatischesLernen lernParam;

	/** Die Daten des Lernparameters. */
	private static PdGanglinienModellAutomatischesLernen.Daten lernDaten;

	/**
	 * Führt Vorbereitungen durch, die einmalig vor dem Testfall notwendig sind.
	 * <ul>
	 * <li>Der Messquerschnitt für den Test wird bestimmt.</li>
	 * <li>Die Testereignisse und -ereignistypen werden angelegt.</li>
	 * <li>Die historischen Ganglinien zu Testbeginn werden angelegt.</li>
	 * <li>Die für den Test benötigten Archivdaten werden ins Archiv
	 * geschrieben.</li>
	 * </ul>
	 *
	 * @throws Exception
	 *             bei einem Fehler.
	 */
	@BeforeClass
	public static void beforeClass() throws Exception {
		final ObjektFactory factory;
		final ClientDavConnection verbindung;
		final Configuration konfig;
		final Level level;

		konfig = Configuration.getConfiguration();
		level = Level.parse(konfig.getString("debug.level"));
		// LoggerTools.setLogggerLevel(level);
		log.setLevel(level);

		log.info("Stelle Verbindung zum Datenverteiler her ...");
		verbindung = new ClientDavConnection();
		verbindung.getClientDavParameters()
				.setDavCommunicationAddress(konfig.getString("dav.host"));
		verbindung.getClientDavParameters()
				.setDavCommunicationSubAddress(konfig.getInt("dav.port"));
		verbindung.connect();
		verbindung.login(konfig.getString("dav.benutzer"),
				konfig.getString("dav.kennwort"));

		factory = ObjektFactory.getInstanz();
		factory.setVerbindung(verbindung);
		factory.registerStandardFactories();
		factory.registerFactory(new GanglinienobjektFactory());

		mq = (MessQuerschnittAllgemein) factory
				.getModellobjekt(konfig.getString("objekt.mq"));
		assertNotNull("Der Testmessquerschnitt existiert nicht.", mq);

		// Archivdaten anlegen
		if (konfig.getBoolean("archivdatenSchreiben")) {
			final ArchivdatenFactory archivdatenFactory;

			archivdatenFactory = new ArchivdatenFactory(mq,
					"jdbc:hsqldb:res:de/bsvrz/iav/gllernen/gllernen/prueffall8;shutdown=true",
					"sa", "");
			archivdatenFactory.anlegenArchivdaten();
		} else {
			log.info("Überspringe anlegen der Archivdaten.");
		}

		// Ereignisse und Ereignistypen anlegen
		ereignisFactory = new EreignisFactory(
				"jdbc:hsqldb:res:de/bsvrz/iav/gllernen/gllernen/prueffall8;shutdown=true",
				"sa", "");
		if (konfig.getBoolean("reset")) {
			ereignisFactory.aufraeumen();
			factory.getVerbindung().sleep(3 * Constants.MILLIS_PER_SECOND);
		}
		ereignisFactory.anlegenEreignistypen();
		factory.getVerbindung().sleep(3 * Constants.MILLIS_PER_SECOND);
		ereignisFactory.anlegenEreignisse();

		// Historische Ganglinien anlegen
		ganglinienFactory = new GanglinienFactory(mq,
				"jdbc:hsqldb:res:de/bsvrz/iav/gllernen/gllernen/prueffall8;shutdown=true",
				"sa", "", konfig.getString("dav.db"));
		if (konfig.getBoolean("reset")) {
			ganglinienFactory.aufraeumen();
			factory.getVerbindung().sleep(3 * Constants.MILLIS_PER_SECOND);
		}
		ganglinienFactory.anlegenGanglinien(GanglinienFactory.ERSTER_TAG);

		// Lernparameter abrufen
		lernParam = factory.getAOE().getParameterDatensatz(
				PdGanglinienModellAutomatischesLernen.class);
		lernParam.anmeldenSender();
		lernDaten = lernParam.abrufenDatum();

		factory.getVerbindung().sleep(10 * Constants.MILLIS_PER_SECOND);
	}

	/**
	 * Testet das Lernen von Ganglinien. Der Test erfolgt in vier Schritten bzw.
	 * "Tagen". An ersten drei Tagen werden absolute Ganglinien gelernt, am
	 * vierten eine relative.
	 * <ol>
	 * <li>Am ersten Tag wird für den Ereignistyp <em>Ostermontag</em> gelernt.
	 * </li>
	 * <li>Am zweiten Tag ist ein <em>Montagsfeiertag</em>, es wird jedoch für
	 * den Ereignistyp <em>Montag</em> gelernt.</li>
	 * <li>Am dritten Tag wird für den Ereignistyp <em>Montag</em> gelernt.</li>
	 * <li>Am vierten Tag wird für Sportereignis gelernt</li>
	 * </ol>
	 * Die Testkonfiguration:
	 * <ul>
	 * <li>Tag 1: Es ist <em>Ostermontag</em>, <em>Montagsfeiertag</em> und
	 * <em>Montag</em>, der <em>Ostermontag</em> hat die höchste Priorität.</li>
	 * <li>Tag 2: Es ist <em>Montagsfeiertag</em> und <em>Montag</em>, der
	 * <em>Montagsfeiertag</em> hat die höchste Priorität. In der
	 * Ausschlussliste von <em>Montagsfeiertag</em> steht <em>Montag</em>.</li>
	 * <li>Tag 3: Es ist nur <em>Montag</em>.</li>
	 * <li>Tag 4: Es ist <em>Dienstag</em> und zusätzliche liegen die relativen
	 * Ereignisse <em>Sportereignis</em> und <em>Unfall</em> an. Die beiden
	 * Ereignisse überlappen sich, <em>Sportereignis</em> hat die höhere
	 * Priorität.</li>
	 * </ul>
	 * Der erste Testtag ist der Tag heute vor 9 Tagen, der letzte Testtag ist
	 * heute vor 6 Tagen. Für jeden Ereignistyp liegen drei historische
	 * Ganglinien mit je fünf Stützstellen (0, 6, 12, 18 und 24 Uhr) vor.
	 * <p>
	 * Die Ganglinien der ersten drei Tage werden jeweils durch einen B-Spline
	 * der Ordnung 5 approximiert, die Ganglinien am vierten Tag je durch einen
	 * Cubic-Spline. Die relativen Ganglinien sind multiplikativ.
	 * <p>
	 * Für jeden der vier Testtage liegt je eine Archivganglinie vor.
	 *
	 * @throws Exception
	 *             bei einem Fehler.
	 */
	@SuppressWarnings("static-method")
	@Test
	public void testLerneGanglinien() throws Exception {
		assertEquals(
				"Das Mindestdatenalter im Lernparameter zu Beginn des Tests ist falsch. Die gefordern Prüfvoraussetzungen laut Prüfprozedur sind nicht erfüllt.",
				10 * Constants.MILLIS_PER_DAY,
				lernDaten.getDatenMindestalter());

		final List<GanglinieMQ> ganglinienSoll_9 = ganglinienFactory
				.getGanglinienTestdaten(-9);
		final List<GanglinieMQ> ganglinienSoll_8 = ganglinienFactory
				.getGanglinienTestdaten(-8);
		final List<GanglinieMQ> ganglinienSoll_7 = ganglinienFactory
				.getGanglinienTestdaten(-7);
		final List<GanglinieMQ> ganglinienSoll_6 = ganglinienFactory
				.getGanglinienTestdaten(-6);
		final List<GanglinieMQ> ganglinienSoll_5 = ganglinienFactory
				.getGanglinienTestdaten(-5);

		final List<GanglinieMQ> ganglinienIst_9 = GlSpeicher.getInstanz()
				.read(mq.getPid());
		assertEquals("Die Ausgangsdaten sind falsch.", ganglinienSoll_9,
				ganglinienIst_9);

		log.info("Stoße das Lernen an (1/4) ...");
		lernDaten.setAktualisierungsintervall(new CronPattern("* * * * *"));
		lernDaten.setDatenMindestalter(9 * Constants.MILLIS_PER_DAY);
		lernDaten.setMaxVergleichsAbstand(100);
		lernParam.sendeDaten(lernDaten);

		log.info("Warte auf Daten (1/4) ...");
		GregorianCalendar cal = new GregorianCalendar();
		cal.set(Calendar.MILLISECOND, 0);
		cal.set(Calendar.SECOND, 0);
		cal.add(Calendar.MINUTE, 1);
		cal.add(Calendar.SECOND, 45);
		Thread.sleep(cal.getTimeInMillis() - System.currentTimeMillis());
		final List<GanglinieMQ> ganglinienIst_8 = GlSpeicher.getInstanz()
				.read(mq.getPid());
		checkGanglinien("Am ersten Tag wurde falsch gelernt.", ganglinienSoll_8,
				ganglinienIst_8);

		log.info("Stoße das Lernen an (2/4) ...");
		lernDaten.setAktualisierungsintervall(new CronPattern("* * * * *"));
		lernDaten.setDatenMindestalter(8 * Constants.MILLIS_PER_DAY);
		lernDaten.setMaxVergleichsAbstand(100);
		lernParam.sendeDaten(lernDaten);

		log.info("Warte auf Daten (2/4) ...");
		cal = new GregorianCalendar();
		cal.set(Calendar.MILLISECOND, 0);
		cal.set(Calendar.SECOND, 0);
		cal.add(Calendar.MINUTE, 1);
		cal.add(Calendar.SECOND, 45);
		Thread.sleep(cal.getTimeInMillis() - System.currentTimeMillis());
		final List<GanglinieMQ> ganglinienIst_7 = GlSpeicher.getInstanz()
				.read(mq.getPid());
		checkGanglinien("Am zweiten Tag wurde falsch gelernt.",
				ganglinienSoll_7, ganglinienIst_7);

		log.info("Stoße das Lernen an (3/4) ...");
		lernDaten.setAktualisierungsintervall(new CronPattern("* * * * *"));
		lernDaten.setDatenMindestalter(7 * Constants.MILLIS_PER_DAY);
		lernDaten.setMaxVergleichsAbstand(100);
		lernParam.sendeDaten(lernDaten);

		log.info("Warte auf Daten (3/4) ...");
		cal = new GregorianCalendar();
		cal.set(Calendar.MILLISECOND, 0);
		cal.set(Calendar.SECOND, 0);
		cal.add(Calendar.MINUTE, 1);
		cal.add(Calendar.SECOND, 45);
		Thread.sleep(cal.getTimeInMillis() - System.currentTimeMillis());
		final List<GanglinieMQ> ganglinienIst_6 = GlSpeicher.getInstanz()
				.read(mq.getPid());
		checkGanglinien("Am dritten Tag wurde falsch gelernt.",
				ganglinienSoll_6, ganglinienIst_6);

		log.info("Stoße das Lernen an (4/4) ...");
		lernDaten.setAktualisierungsintervall(new CronPattern("* * * * *"));
		lernDaten.setDatenMindestalter(6 * Constants.MILLIS_PER_DAY);
		lernDaten.setMaxVergleichsAbstand(100);
		lernParam.sendeDaten(lernDaten);

		log.info("Warte auf Daten (4/4) ...");
		cal = new GregorianCalendar();
		cal.set(Calendar.MILLISECOND, 0);
		cal.set(Calendar.SECOND, 0);
		cal.add(Calendar.MINUTE, 1);
		cal.add(Calendar.SECOND, 45);
		Thread.sleep(cal.getTimeInMillis() - System.currentTimeMillis());
		final List<GanglinieMQ> ganglinienIst_5 = GlSpeicher.getInstanz()
				.read(mq.getPid());
		checkGanglinien("Am vierten Tag wurde falsch gelernt.",
				ganglinienSoll_5, ganglinienIst_5);
	}

	private static void checkGanglinien(final String error,
			final List<GanglinieMQ> ganglinienSoll,
			final List<GanglinieMQ> ganglinienIst) {
		Assert.assertEquals(error + " Die Anzahl der Ganglinien ist falsch.",
				ganglinienSoll.size(), ganglinienIst.size());

		for (int i = 0; i < ganglinienSoll.size(); i++) {
			Assert.assertEquals(
					error + " Die Ganglinien mit dem Index " + i
							+ " ist falsch.",
					ganglinienSoll.get(i), ganglinienIst.get(i));
		}
	}

	/**
	 * Räumt nach dem Test auf.
	 * <ul>
	 * <li>Die Testereignisse und -ereignistypen werden wieder gelöscht.</li>
	 * </ul>
	 *
	 * @throws Exception
	 *             bei einem Fehler.
	 */
	@AfterClass
	public static void afterClass() throws Exception {
		lernDaten.setAktualisierungsintervall(new CronPattern("* * * * *"));
		lernDaten.setDatenMindestalter(10 * Constants.MILLIS_PER_DAY);
		lernDaten.setMaxVergleichsAbstand(100);
		lernParam.sendeDaten(lernDaten);
		ereignisFactory.aufraeumen();
		ganglinienFactory.aufraeumen();
	}

}
